import logging
from abc import ABCMeta, abstractmethod
from datetime import datetime


class Video:
    def __init__(self,
                 video_id: str,
                 channel_id: str,
                 timestamp: datetime,
                 title: str,
                 channel_name: str,
                 filename: str=None):
        self.video_id = video_id
        self.channel_id = channel_id
        self.timestamp = timestamp
        self.title = title
        self.channel_name = channel_name
        self.filename = filename

    def __str__(self):
        return '{}\t{}\t{}\t{}\t{}\t{}'.format(
            self.video_id,
            self.channel_id,
            self.timestamp,
            self.title,
            self.channel_name,
            self.filename
        )

    def __repr__(self):
        return self.__str__()


class Recordings(metaclass=ABCMeta):
    @abstractmethod
    def update(self, logger: logging.Logger):
        pass

    @abstractmethod
    def is_recording_active(self, recording_id: str) -> bool:
        pass

    @abstractmethod
    def create_recording(self, stream: Video, logger: logging.Logger):
        pass


class RecorderMessage:
    def __init__(self, recorder_to_shutdown: str):
        self.recorder_to_shutdown = recorder_to_shutdown


class Context:
    def __init__(self, config, logger: logging.Logger, recordings_controller: Recordings):
        self.config = config
        self.logger = logger
        self.service = None
        self.recordings = recordings_controller



import os
import sqlite3
from contextlib import contextmanager
from typing import Iterator, List

from ytarchiver.common import Video, Context

STORAGE_FILE = 'storage.sqlite'


class Storage:
    def __init__(self, output_directory: str, initialise: bool=True):
        path = os.path.join(output_directory, STORAGE_FILE)
        storage_exist = os.path.isfile(path)

        self.output_directory = output_directory
        self.connection = sqlite3.connect(path)

        if initialise and not storage_exist:
            self._initialise_schema()

    def close(self):
        self.connection.commit()
        self.connection.close()

    def list_videos(self) -> Iterator[Video]:
        cur = self.connection.cursor()
        cur.execute('SELECT video_id, channel_id, timestamp, title, channel_name, filename FROM VIDEOS')
        for columns in cur.fetchall():
            yield Video(*columns)

    def list_livestreams(self) -> Iterator[Video]:
        cur = self.connection.cursor()
        cur.execute('SELECT video_id, channel_id, timestamp, title, channel_name, filename FROM LIVESTREAMS')
        for columns in cur.fetchall():
            yield Video(*columns)

    def video_exist(self, video_id: str):
        cur = self.connection.cursor()
        cur.execute('SELECT 1 FROM VIDEOS WHERE video_id=?', (video_id,))
        return len(list(cur.fetchall())) > 0

    def add_video(self, entry: 'Video'):
        cur = self.connection.cursor()
        cur.execute(
            'INSERT INTO VIDEOS(video_id, channel_id, timestamp, title, channel_name, filename) '
            'VALUES (?, ?, ?, ?, ?, ?)',
            (entry.video_id, entry.channel_id, entry.timestamp, entry.title, entry.channel_name, entry.filename)
        )

    def add_livestream(self, entry: 'Video'):
        cur = self.connection.cursor()
        cur.execute(
            'INSERT INTO LIVESTREAMS(video_id, channel_id, timestamp, title, channel_name, filename) '
            'VALUES (?, ?, ?, ?, ?, ?)',
            (entry.video_id, entry.channel_id, entry.timestamp, entry.title, entry.channel_name, entry.filename)
        )
        self.commit()

    def commit(self):
        self.connection.commit()

    def _initialise_schema(self):
        cur = self.connection.cursor()
        cur.execute(
            'CREATE TABLE VIDEOS('
            'video_id VARCHAR(16) PRIMARY KEY, '
            'channel_id VARCHAR(32), '
            'timestamp DATETIME, '
            'title TEXT, '
            'channel_name TEXT, '
            'filename TEXT'
            ')'
        )
        cur.execute(
            'CREATE TABLE LIVESTREAMS('
            'id INTEGER PRIMARY KEY AUTOINCREMENT,'
            'video_id VARCHAR(16), '
            'channel_id VARCHAR(32), '
            'timestamp DATETIME, '
            'title TEXT, '
            'channel_name TEXT, '
            'filename TEXT'
            ')'
        )
        self.connection.commit()


@contextmanager
def open_storage(directory: str, initialise: bool=True) -> Storage:
    s = Storage(directory, initialise=initialise)
    yield s
    s.close()


def get_saved_content(context: Context) -> List[Video]:
    storage_path = os.path.join(context.config.output_dir, STORAGE_FILE)
    if not os.path.isfile(storage_path):
        context.logger.warning('storage file does not exist: ' + storage_path)
        return []

    with open_storage(context.config.output_dir, initialise=False) as s:
        return list(s.list_videos()) + list(s.list_livestreams())



import argparse

DEFAULT_REFRESH_TIME_SEC = 5 * 60  # 5 min
DEFAULT_OUTPUT_DIRECTORY = './out'
DEFAULT_ARCHIVE_ALL = False
DEFAULT_MONITOR_LIVESTREAMS = False


def parse_command_line():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-l', '--list',
        dest='do_list',
        help='do not listen for videos but rather list already saved ones',
        default=False,
        action='store_true'
    )
    parser.add_argument(
        '-t', '--time',
        dest='refresh_time',
        help='time between consecutive lookups for videos (in seconds)',
        default=DEFAULT_REFRESH_TIME_SEC,
        type=int
    )
    parser.add_argument(
        '-o', '--output',
        dest='output_dir',
        help='directory to store videos and the database, default: ' + DEFAULT_OUTPUT_DIRECTORY,
        default=DEFAULT_OUTPUT_DIRECTORY,
        type=str
    )
    parser.add_argument(
        '-a', '--all',
        dest='archive_all',
        help='archive all videos from past, default action is to save only new videos',
        default=DEFAULT_ARCHIVE_ALL,
        action='store_true'
    )
    parser.add_argument(
        '-s', '--streams',
        dest='monitor_livestreams',
        help='also monitor active livestreams and record all of them',
        default=DEFAULT_MONITOR_LIVESTREAMS,
        action='store_true'
    )
    parser.add_argument(
        '-k', '--key',
        dest='api_key',
        help='Google API key',
        default=None,
        type=str
    )
    parser.add_argument(
        '-m', '--monitor',
        dest='channels_list',
        help='list of channels to monitor, use channel IDs split by space such as: UC8e8Ag9Or9ra4J6jl4qjfEg ...',
        nargs='+',
        default=[]
    )

    config = parser.parse_args()
    return config



import errno
import logging
import os
import sys
import time
from datetime import datetime

from ytarchiver.args import parse_command_line
from ytarchiver.api import prepare_context
from ytarchiver.common import Context
from ytarchiver.lookup import lookup
from ytarchiver.recording import MultiprocessRecordingsController
from ytarchiver.storage import get_saved_content


def main():
    config = parse_command_line()
    logger = create_logger()
    context = Context(config, logger, recordings_controller=MultiprocessRecordingsController())

    if config.do_list:
        list_videos(context)
    else:
        start_listening(context)


def create_logger():
    logging.basicConfig(level=logging.DEBUG, stream=sys.stderr, format='%(message)s')
    logging.getLogger('googleapiclient.discovery').setLevel(logging.ERROR)
    logging.getLogger('urllib3.connectionpool').setLevel(logging.ERROR)
    logger = logging.getLogger('ytarchiver')
    return logger


def list_videos(context: Context):
    context.logger.info('Video ID\tChannel ID\tTimestamp\tTitle\tChannel Name\tFilename')
    for entry in get_saved_content(context):
        context.logger.info(entry)


def start_listening(context: Context):
    ensure_dir_exist(context.config.output_dir, context.logger)
    if len(context.config.channels_list) == 0:
        context.logger.error('channels list cannot be empty, use -m option to specify at least one channel id')
        sys.exit(1)

    prepare_context(context)

    context.logger.debug('daemon started successfully')

    try:
        _trigger_lookup(context, first_run=True)

        while True:
            time.sleep(context.config.refresh_time)
            _trigger_lookup(context)
    except KeyboardInterrupt:
        context.logger.debug('interrupted with ctrl+c')


def _trigger_lookup(context, first_run=False):
    now = datetime.now()
    context.logger.debug('[{}] triggering new lookup...'.format(now.strftime('%Y-%m-%d %H:%M:%S')))
    lookup(context, first_run)


def ensure_dir_exist(path: str, logger: logging.Logger):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            logger.error(e)
            sys.exit(1)


if __name__ == '__main__':
    main()



import os
import logging

import streamlink
from pytube import YouTube
from pytube.helpers import safe_filename

from ytarchiver.common import Video, Context

YOUTUBE_URL_PREFIX = 'https://www.youtube.com/watch?v='
SUPPORTED_LIVESTREAM_RESOLUTIONS = ['720p', '480p', '360p', '240p', '144p']
LIVESTREAM_CHUNK_SIZE = 8192


def record_livestream(livestream: Video, logger: logging.Logger):
    url = YOUTUBE_URL_PREFIX + livestream.video_id
    available_streams = streamlink.api.streams(url)

    best_resolution = ''
    for resolution in SUPPORTED_LIVESTREAM_RESOLUTIONS:
        if resolution in available_streams:
            best_resolution = resolution
            break
    if best_resolution == '':
        logger.error('no supported resolution found for "{}"'.format(livestream.title))
        return

    stream = available_streams[best_resolution]

    logging.error('recording {}:{} stream of "{}"'.format(stream.shortname(), best_resolution, livestream.title))
    with open(livestream.filename, 'wb') as out:
        with stream.open() as handle:
            while True:
                buffer = handle.read(LIVESTREAM_CHUNK_SIZE)
                out.write(buffer)
                out.flush()


def download_video(context: Context, video: Video):
    download_link = YOUTUBE_URL_PREFIX + video.video_id
    total_size = 0

    def on_progress(stream, chunk, handle, bytes_remaining):
        context.logger.debug('downloading... {}/{}'.format(total_size - bytes_remaining, total_size))

    def on_complete(stream, handle):
        context.logger.debug('download complete')

    yt = YouTube(download_link)
    yt.register_on_complete_callback(on_complete)
    yt.register_on_progress_callback(on_progress)
    streams = yt.streams.all()
    stream = _choose_best_stream(streams)
    total_size = stream.filesize
    filename = _sanitize_filename(video.timestamp + ' ' + video.title)

    context.logger.info('downloading "{}"'.format(video.title))
    stream.download(
        output_path=context.config.output_dir,
        filename=filename
    )

    return os.path.join(context.config.output_dir, '{}.{}'.format(filename, stream.subtype))


def _choose_best_stream(streams):
    best_stream = None
    for stream in streams:
        if stream.includes_audio_track and stream.includes_video_track:
            if best_stream is None:
                best_stream = stream

            higher_resolution = stream.resolution > best_stream.resolution
            equal_resolution = stream.resolution == best_stream.resolution
            better_format = 'mp4' in stream.mime_type and 'mp4' not in best_stream.mime_type
            if better_format and (equal_resolution or higher_resolution):
                best_stream = stream
            if higher_resolution:
                best_stream = stream

    return best_stream


def _sanitize_filename(filename):
    filename = safe_filename(filename)
    filename = filename.replace(' ', '_')
    return filename.encode('ascii', 'ignore').decode('ascii')



from typing import Iterator, Optional

from googleapiclient.discovery import build

from ytarchiver.common import Context, Video

YOUTUBE_API_SERVICE_NAME = "youtube"
YOUTUBE_API_VERSION = "v3"


def prepare_context(context: Context):
    context.service = build(
        YOUTUBE_API_SERVICE_NAME,
        YOUTUBE_API_VERSION,
        developerKey=context.config.api_key,
        cache_discovery=False
    )


def fetch_channel_livestream(context: Context, channel_id: str) -> Optional[Video]:
    live_streams = context.service.search().list(
        part="id,snippet",
        channelId=channel_id,
        type='video',
        eventType='live'
    ).execute()

    for stream in live_streams['items']:
        return Video(
            video_id=stream['id']['videoId'],
            channel_id=channel_id,
            timestamp=stream['snippet']['publishedAt'],
            title=stream['snippet']['title'],
            channel_name=stream['snippet']['channelTitle']
        )

    return None


def find_channel_uploaded_videos(context: Context, channel_id: str) -> Iterator[Video]:
    results = context.service.channels().list(
        part="snippet,contentDetails",
        id=channel_id
    ).execute()

    for channel in results['items']:
        uploads_playlist_id = channel['contentDetails']['relatedPlaylists']['uploads']
        next_page_token = ''

        while next_page_token is not None:
            playlistitems_response = context.service.playlistItems().list(
                playlistId=uploads_playlist_id,
                part='snippet',
                maxResults=50,
                pageToken=next_page_token
            ).execute()

            for playlist_item in playlistitems_response['items']:
                upload = playlist_item['snippet']
                yield Video(
                    video_id=upload['resourceId']['videoId'],
                    channel_id=upload['channelId'],
                    timestamp=upload['publishedAt'],
                    title=upload['title'],
                    channel_name=upload['channelTitle']
                )

            next_page_token = playlistitems_response.get('nextPageToken')



from distutils.core import setup
from setuptools import find_packages

with open('README.md') as f:
    readme = f.read()

with open('LICENSE.txt') as f:
    license = f.read()


setup(
    name='yt-archiver',
    version='0.3.5',
    description='Daemon for monitoring YouTube channels for new videos. Offers automatic backup of uploaded content',
    long_description=readme,
    license=license,
    author='Michal Korman',
    author_email='michal.korman@icloud.com',
    url='https://github.com/mkorman9/yt-archiver',
    packages=find_packages(),
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS',
        'Operating System :: Microsoft',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python',
        'Environment :: Console',
    ],
    entry_points={
        'console_scripts': [
            'ytarchiver = ytarchiver.main:main'
        ]
    },
    install_requires=['pytube', 'streamlink', 'google-api-python-client'],
)



import os
from datetime import datetime

from googleapiclient.errors import HttpError

from ytarchiver.api import find_channel_uploaded_videos, fetch_channel_livestream
from ytarchiver.common import Context, Video
from ytarchiver.download import download_video, _sanitize_filename
from ytarchiver.storage import open_storage, Storage


class Statistics:
    def __init__(self):
        self.total_videos = 0
        self.new_videos = 0
        self.active_livestreams = 0

    def notify_video(self, new: bool=False):
        self.total_videos += 1
        if new:
            self.new_videos += 1

    def notify_active_livestream(self):
        self.active_livestreams += 1


def lookup(context: Context, is_first_run: bool):
    statistics = Statistics()

    context.recordings.update(context.logger)

    with open_storage(context.config.output_dir) as storage:
        for channel_id in context.config.channels_list:
            _fetch_channel_content(context, channel_id, storage, statistics, is_first_run)

    context.logger.info(
        'total videos: {}, new: {}, active livestreams {}'.format(
            statistics.total_videos,
            statistics.new_videos,
            statistics.active_livestreams if context.config.monitor_livestreams else 'N/A'
        )
    )


def _fetch_channel_content(context: Context, channel_id: str, storage: Storage, statistics, is_first_run: bool=False):
    try:
        _check_for_livestreams(context, channel_id, statistics, storage)
        _check_for_videos(context, channel_id, is_first_run, statistics, storage)
    except ConnectionError as e:
        context.logger.error('connection to API has failed, skipping lookup')
        context.logger.error(e)
        return
    except HttpError as e:
        context.logger.error('API call has failed, skipping lookup')
        context.logger.error(e)
        return
    finally:
        storage.commit()


def _check_for_livestreams(context: Context, channel_id: str, statistics: Statistics, storage: Storage):
    if not context.config.monitor_livestreams:
        return

    livestream = fetch_channel_livestream(context, channel_id)
    if livestream is not None:
        if not context.recordings.is_recording_active(livestream.video_id):
            livestream.filename = _generate_livestream_filename(context, livestream)
            context.logger.info('new livestream "{}"'.format(livestream.title))
            context.recordings.create_recording(livestream, context.logger)
            storage.add_livestream(livestream)
        statistics.notify_active_livestream()


def _check_for_videos(context: Context, channel_id: str, is_first_run: bool, statistics: Statistics, storage: Storage):
    for video in find_channel_uploaded_videos(context, channel_id):
        video_not_registered = not storage.video_exist(video.video_id)
        if video_not_registered:
            _register_video(context, storage, video, is_first_run)
        statistics.notify_video(new=video_not_registered)


def _register_video(context: Context, storage: Storage, video: Video, is_first_run: bool=False):
    should_be_downloaded = not is_first_run or context.config.archive_all

    context.logger.info('new video "{}"'.format(video.title))

    if should_be_downloaded:
        file_path = download_video(context, video)
        video.filename = file_path

    storage.add_video(video)

    if should_be_downloaded:
        storage.commit()


def _generate_livestream_filename(context: Context, livestream: Video):
    now = datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f')
    filename = '{}_{}.{}'.format(now, _sanitize_filename(livestream.title), 'ts')
    return os.path.join(context.config.output_dir, filename)



import logging
from multiprocessing import Queue, Process
from queue import Empty
from typing import Iterator

from ytarchiver.common import Recordings, RecorderMessage, Video
from ytarchiver.download import record_livestream


class MultiprocessRecordingsController(Recordings):
    def __init__(self):
        self.active_recordings = {}
        self.recorders_queue = MultiprocessRecordersQueue()

    def update(self, logger: logging.Logger):
        for message in self.recorders_queue.check_for_messages():
            del self.active_recordings[message.recorder_to_shutdown]

    def is_recording_active(self, recording_id: str) -> bool:
        return recording_id in self.active_recordings

    def create_recording(self, stream: Video, logger: logging.Logger):
        self.active_recordings[stream.video_id] = True  # replace boolean with any other information about recording

        process = Process(
            name='ytarchiver-recorder',
            target=_recorder_task,
            args=(stream, self.recorders_queue, logger)
        )
        process.start()


class MultiprocessRecordersQueue:
    def __init__(self):
        self._queue = Queue()

    def check_for_messages(self) -> Iterator[RecorderMessage]:
        try:
            message = RecorderMessage('')
            while message is not None:
                message = self._queue.get(block=False)
                yield message
        except Empty:
            pass

    def send_message(self, message: RecorderMessage):
        self._queue.put(message, block=False)


def _recorder_task(stream: Video, recorders_queue: MultiprocessRecordersQueue, logger: logging.Logger):
    try:
        record_livestream(stream, logger)
    except (IOError, EOFError):
        logger.error('stream "{}" finished unexpectedly'.format(stream.title))
    except Exception as e:
        logger.error('error while recording stream "{}"'.format(stream.title))
        logger.error(e)
    finally:
        logger.error('recording of stream "{}" has ended'.format(stream.title))
        recorders_queue.send_message(
            RecorderMessage(recorder_to_shutdown=stream.video_id)
        )
